#!/usr/bin/env python3

import mxnet as mx
import os
import cv2
import argparse
import operator
from mxnet import image
from mxnet import nd
from mxnet import init
from mxnet.gluon.model_zoo.vision import vgg19_bn

class PredictionResult:
    def __init__(self, current_label, results, prediction_label = None):
        self.current_frame_label = current_label
        self.all_results = results
        self.long_term_prediction_label = prediction_label

class Predictor:
    def __init__(self, context, net):
        self.net = net
        self.context = context
        self.min_results_count= 20
        self.last_results = []
        self.labels = ['dry', 'snow', 'wet']

    def _transform(self, data):
        test_augs = [
            image.ResizeAug(224),
            image.CenterCropAug((224, 224))
        ]
        data = data.astype('float32')
        for aug in test_augs:
            data = aug(data)
        data = nd.transpose(data, (2,0,1))
        return data

    def _get_current_label_and_results(self, out):
        pred = int(nd.argmax(out, axis=1).asscalar())
        results = {}
        for i in range(len(self.labels)):
            results[self.labels[i]] = out[0][i].asscalar()
        return self.labels[pred], results
    
    def _add_new_result(self, label):
        self.last_results.append(label)
        if len(self.last_results) > self.min_results_count:
            self.last_results.pop(0)
    
    def _get_estimation(self):
        if len(self.last_results) < self.min_results_count:
            return None
        _results = {}
        for label in self.labels:
            _results[label] = 0
        
        points = 1
        for label in self.last_results:
            _results[label] += points
            points += 1
        
        return max(_results.items(), key=operator.itemgetter(1))[0]
        

    def predict_road_state(self, img):
        data = self._transform(img)
        data = data.as_in_context(self.context)
        data = data.expand_dims(axis=0)
        data = image.color_normalize(data/255,
                            mean=nd.array([0.485, 0.456, 0.406], self.context).reshape((1,3,1,1)),
                            std=nd.array([0.229, 0.224, 0.225], self.context).reshape((1,3,1,1)))
        out = self.net(data)
        label, results = self._get_current_label_and_results(out)
        self._add_new_result(label)
        
        result = PredictionResult(label, results, self._get_estimation())
        return result

class VideoGrabber:
    def __init__(self, args):
        if args.filepath is not None:
            self.video_grabber = cv2.VideoCapture(args.filepath)
        else:
            self.video_grabber = cv2.VideoCapture(int(args.cam))
        self.retrived_frames = 0

    def __del__(self): 
        self.video_grabber.release()
    
    def is_frame_available(self):
        return (self.video_grabber.grab() > 0)
    
    def get_frame(self):
        ret, frame = self.video_grabber.retrieve()
        self.retrived_frames += 1
        return frame
    
    def skip_frames(self, skip_by):
        for j in range(skip_by):
            self.video_grabber.grab()
    
    def get_retrived_frames_count(self):
        return self.retrived_frames

class Utils:
    def mx_array_to_opencv_mat(self, data, transposed=False):
        result = data
        if transposed == True:
            result = nd.transpose(result, (1,2,0))
        return cv2.cvtColor(result.astype('uint8').asnumpy(), cv2.COLOR_RGB2BGR)

    def opencv_mat_to_array(self, frame):
        return mx.nd.array(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))


def prepare_nn(context, params_snapshot_file):
    pretrained_net = vgg19_bn(pretrained=True, ctx=context)

    net = vgg19_bn(classes=3, ctx=context)
    net.features = pretrained_net.features
    net.output.initialize(init.Xavier(), ctx=context)

    net.hybridize()
    net.load_parameters(filename=params_snapshot_file, ctx=context)
    return net

def parse_cmd_line_args():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--filepath", help="path of video file to analyze")
    group.add_argument("--cam", type=int, help="video cam id to analyze frames from")
    parser.add_argument("--check_ratio", default=10, type=int, help="(for video file only) - how often send frames to prediction algorithm. e.g. '5' means every 5th frame will be checked.")
    parser.add_argument("-v", action="store_true", help="open window with grabbed frames (press 'q' to quit)")
    group2 = parser.add_mutually_exclusive_group(required=True)
    group2.add_argument("--cpu", action="store_true", help="run processing on cpu")
    group2.add_argument("--gpu", action="store_true", help="run processing on gpu")
    args = parser.parse_args()
    return args

def show_frame(frame, label):
    cv2.putText(frame, label, (20,50), fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=1, color=(255, 0, 0), thickness=2)
    cv2.imshow('Road state',frame)

def main():
    args = parse_cmd_line_args()
    if args.gpu == True:
        global_context = mx.gpu()
    else:
        global_context = mx.cpu()
    predictor = Predictor(
        global_context, 
        prepare_nn(global_context, './checkpoints/vgg19_bn_15.params')
    )
    video_grabber = VideoGrabber(args)

    current_label = 'unknown'
    print("Current prediction: " + current_label)
    while video_grabber.is_frame_available():
        frame = video_grabber.get_frame()
        if video_grabber.get_retrived_frames_count() % args.check_ratio == 0:
            result = predictor.predict_road_state(Utils().opencv_mat_to_array(frame))
            if result.long_term_prediction_label is not None:
                if result.long_term_prediction_label != current_label:
                    print("Current prediction: " + result.long_term_prediction_label)
                current_label = result.long_term_prediction_label
        if args.v == True:
            show_frame(frame, current_label)
            key = cv2.waitKey(1) & 0xFF
            if key == ord('q'):
                break
            elif key == ord('n'):
                video_grabber.skip_frames(100)

    cv2.destroyAllWindows()

main()